@@ -52,8 +52,10 @@ static long hpage_size;
 
 static void cacheflush(void *p)
 {
-#ifdef __powerpc__
+#if defined(__powerpc__)
 	asm volatile("dcbst 0,%0; sync; icbi 0,%0; isync" : : "r"(p));
+#elif defined(__arm__) || defined(__aarch64__)
+	__clear_cache(p, p + COPY_SIZE);
 #endif
 }
 
@@ -85,8 +87,9 @@ static void *sig_expected;
 static void sig_handler(int signum, siginfo_t *si, void *uc)
 {
 #if defined(__powerpc__) || defined(__powerpc64__) || defined(__ia64__) || \
-    defined(__s390__) || defined(__s390x__) || defined(__sparc__)
-	/* On powerpc and ia64 and s390, 0 bytes are an illegal
+    defined(__s390__) || defined(__s390x__) || defined(__sparc__) || \
+    defined(__aarch64__)
+	/* On powerpc, ia64, s390 and Aarch64, 0 bytes are an illegal
 	 * instruction, so, if the icache is cleared properly, we SIGILL
 	 * as soon as we jump into the cleared page */
 	if (signum == SIGILL) {
@@ -97,7 +100,7 @@ static void sig_handler(int signum, siginfo_t *si, void *uc)
 		}
 		FAIL("SIGILL somewhere unexpected");
 	}
-#elif defined(__i386__) || defined(__x86_64__)
+#elif defined(__i386__) || defined(__x86_64__) || defined(__arm__)
 	/* On x86, zero bytes form a valid instruction:
 	 *	add %al,(%eax)		(i386)
 	 * or	add %al,(%rax)		(x86_64)
@@ -109,7 +112,14 @@ static void sig_handler(int signum, siginfo_t *si, void *uc)
 	 * on the second, truncated page.  If [ER]AX does not contain
 	 * a valid pointer, we will SEGV on the first instruction in
 	 * the cleared page.  We check for both possibilities
-	 * below. */
+	 * below.
+	 *
+	 * On 32 bit ARM, zero bytes are interpreted as follows:
+	 * 	andeq	r0, r0, r0	(ARM state, 4 bytes)
+	 * 	movs	r0, r0		(Thumb state, 2 bytes)
+	 *
+	 * So, we only expect to run off the end of the huge page and
+	 * generate a SIGBUS. */
 	if (signum == SIGBUS) {
 		verbose_printf("SIGBUS at %p (sig_expected=%p)\n", si->si_addr,
 			       sig_expected);
@@ -120,6 +130,7 @@ static void sig_handler(int signum, siginfo_t *si, void *uc)
 		}
 		FAIL("SIGBUS somewhere unexpected");
 	}
+#if defined(__x86_64__) || defined(__i386__)
 	if (signum == SIGSEGV) {
 #ifdef __x86_64__
 		void *pc = (void *)((ucontext_t *)uc)->uc_mcontext.gregs[REG_RIP];
@@ -134,6 +145,7 @@ static void sig_handler(int signum, siginfo_t *si, void *uc)
 		}
 		FAIL("SIGSEGV somewhere unexpected");
 	}
+#endif
 #else
 #error Need to setup signal conditions for this arch
 #endif
@@ -142,12 +154,13 @@ static void sig_handler(int signum, siginfo_t *si, void *uc)
 static void test_once(int fd)
 {
 	void *p, *q;
+	int dummy;
 
-	ftruncate(fd, 0);
+	dummy = ftruncate(fd, 0);
 
 	if (sigsetjmp(sig_escape, 1)) {
 		sig_expected = NULL;
-		ftruncate(fd, 0);
+		dummy = ftruncate(fd, 0);
 		return;
 	}
 
@@ -156,13 +169,13 @@ static void test_once(int fd)
 	if (p == MAP_FAILED)
 		FAIL("mmap() 1: %s", strerror(errno));
 
-	ftruncate(fd, hpage_size);
+	dummy = ftruncate(fd, hpage_size);
 
 	q = p + hpage_size - COPY_SIZE;
 
 	jumpfunc(1, q);
 
-	ftruncate(fd, 0);
+	dummy = ftruncate(fd, 0);
 	p = mmap(p, hpage_size, PROT_READ|PROT_WRITE|PROT_EXEC,
 		 MAP_SHARED|MAP_FIXED, fd, 0);
 	if (p == MAP_FAILED)
